package eu.asvachin.audiorouter;

import eu.asvachin.audiorouter.gui.AudioRouterGui;
import eu.asvachin.toolbox.properties.Properties;
import eu.asvachin.toolbox.util.Logging;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;

/**
 * Audio Router.
 *
 * @version 2.0 22 August 2012
 * @author Miles Asvachin
 */
public final class Main
{
	private static InputReader reader;
	private static OutputWriter writer;
	
	/**
	 * @param args the command line arguments
	 */
	public static void main(String[] args) throws IOException
	{
		Logging.initialise(false);
		Logger.getLogger(Main.class.getName()).log(Level.INFO, "Started");
		
		//Make sure the program quits when there is an exception, rather than 
		//the default behaviour of only the affected thread being terminated.
		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler()
		{
			@Override
			public void uncaughtException(final Thread t, final Throwable ex)
			{
				Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "Uncaught exception", ex);
				System.exit(1);
			}
		});
		
		javax.swing.SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				final float sampleRate = Properties.getAppProperties().getProperty(float.class, "format", "sampleRate");
				final int resolution = Properties.getAppProperties().getProperty(int.class, "format", "resolution");
				final int channels = Properties.getAppProperties().getProperty(int.class, "format", "channels");
				final AudioFormat format = new AudioFormat(sampleRate, resolution, channels, true, false);
				
				final int bufferSize = Properties.getAppProperties().getProperty(int.class, "bufferSize");
				
				final String inputName = Properties.getAppProperties().getProperty(String.class, "mixers", "input");
				final String outputName = Properties.getAppProperties().getProperty(String.class, "mixers", "output");
				final Mixer.Info input = getMixerFromName(inputName);
				final Mixer.Info output = getMixerFromName(outputName);
				
				new AudioRouterGui(new Connector()
				{
					@Override
					public void connect(final Mixer.Info input, final Mixer.Info output, final AudioFormat format, final int bufferSamples) throws LineUnavailableException
					{
						Main.connect(input, output, format, bufferSamples);
					}

					@Override
					public void disconnect()
					{
						Main.disconnect();
					}
				},
					input,
					output,
					format,
					bufferSize);
			}
		});
	}
	
	public static Mixer.Info getMixerFromName(String name)
	{
		final Mixer.Info[] mixers = AudioSystem.getMixerInfo();
		Mixer.Info matchingMixer = null;
		
		for(final Mixer.Info mixer : mixers)
		{
			if(mixer.getName().equals(name))
			{
				matchingMixer = mixer;
				break;
			}
		}
		
		return matchingMixer;
	}
	
	//@todo The GUI should allow selection of buffer size in ms, and show the calculated size in bytes. Pass it to here in bytes.
	private static void connect(final Mixer.Info input, final Mixer.Info output, final AudioFormat format, final int bufferSamples) throws LineUnavailableException
	{
		try
		{
			final int sampleBytes = format.getSampleSizeInBits() / 8;
			final int bufferBytes = bufferSamples * sampleBytes * format.getChannels();
			final Buffer buffer = new Buffer(bufferBytes);
			reader = InputReader.createReader(input, format, buffer);
			writer = OutputWriter.createWriter(output, format, buffer);
			Logger.getLogger(Main.class.getName()).log(Level.INFO, "Connected {0} to {1}", new Object[]{input.getName(), output.getName()});
		}
		catch(LineUnavailableException | IllegalArgumentException ex)
		{
			Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "Failed to connect " + input.getName() + " to " + output.getName(), ex);
			disconnect();
			throw ex;
		}
	}
	
	private static void disconnect()
	{
		if(writer != null)
		{
			writer.stop();
			writer = null;
		}
		
		if(reader != null)
		{
			reader.stop();
			reader = null;
		}
		
		Logger.getLogger(Main.class.getName()).log(Level.INFO, "Disconnected");
	}
	
	public interface Connector
	{
		public void connect(final Mixer.Info input, final Mixer.Info output, final AudioFormat format, final int bufferSamples) throws LineUnavailableException;
		public void disconnect();
	}
}
